// make HidIn and HidMsg
Hid hi;
HidMsg msg;

// which joystick
0 => int device;
// get from command line
if(me.args()) me.arg(0) => Std.atoi => device;

// open joystick 0, exit on fail
if(!hi.openJoystick(device)) me.exit();

<<< "joystick '" + hi.name() + "' ready", "" >>>;

// set up global variables
0.2 => dac.gain;
200 => float eighth;
0 => int currentChord;
false => int downbeat;
false => int tempoChangeWaiting;
0 => int leftJoyMode;
0 => int rightJoyMode;
false => int leftJoyLocked;
false => int rightJoyLocked;
1 => float freqMultiplier;

[[61, 65, 68, 72], [60, 63, 68, 72], [63, 67, 70, 75], [65, 68, 72, 77]] @=> int arpchordNotes[][];
[37, 32, 39, 41] @=> int bassNotes[];

// instrument effect rack
Gain inGain => Echo inDel => PRCRev inRev => LPF lpf => dac;
0.25 => float inGainDefault => inGain.gain;
135::ms => inDel.delay;
0 => inDel.mix;
0 => inRev.mix;
20000 => lpf.freq;
2 => lpf.Q;

// saw arpeggiator
SawOsc arp => ADSR arpADSR => Gain arpGain => inGain;
0 => arp.gain;
0::ms => arpADSR.attackTime;
2*eighth::ms => arpADSR.decayTime;
1.25 => arpGain.gain;

fun void runArp() {
    while (true) {
        for (0 => int j; j < 4; j++) {
            if (j == 0) {
                true => downbeat;
            } else {
                false => downbeat;
            }
            5::ms => now;
            arpchordNotes[currentChord][j] => Std.mtof => arp.freq;
            arpADSR.keyOn();
            (eighth-5)::ms => now;
        }
    }   
}
spork ~ runArp();

// saw chords
Gain chordGain => inGain;
0.85 => chordGain.gain;

SawOsc chord0 => chordGain;
SawOsc chord1 => chordGain;
SawOsc chord2 => chordGain;
SawOsc chord3 => chordGain;
0 => chord0.gain;
0 => chord1.gain;
0 => chord2.gain;
0 => chord3.gain;

fun void runChord() {
    while (true) {
        5::ms => now;
        arpchordNotes[currentChord][0]-12 => Std.mtof => chord0.freq;
        arpchordNotes[currentChord][1]-12 => Std.mtof => chord1.freq;
        arpchordNotes[currentChord][2]-12 => Std.mtof => chord2.freq;
        arpchordNotes[currentChord][3]-12 => Std.mtof => chord3.freq;
        (eighth-5)::ms => now;
    }
}
spork ~ runChord();

// saw bass
SawOsc bass => inGain;
0 => bass.gain;

fun void runBass() {
    while (true) {
        bassNotes[currentChord] => Std.mtof => bass.freq;
        eighth::ms => now;
    }   
}
spork ~ runBass();

// drum effect rack
Gain drGain => Echo drDel => PRCRev drRev => lpf => dac;
1 => drGain.gain;
135::ms => drDel.delay;
0 => drDel.mix;
0 => drRev.mix;

// kick
SndBuf kick => drGain;
me.dir() + "/sounds/kick.wav" => kick.read;
0 => kick.gain;

fun void runKick() {
    0 => kick.pos;
    1::second => now;
}
fun void kickPattern() { // generative kick patterns
    while (true) {
        5::ms => now;
        (eighth-5)::ms => now;
        Math.randomf() => float pattern;
        if (pattern < 0.7) { // 70% chance of 1
            spork ~ runKick();
        }
        for (0 => int i; i < 3; i++) {
            eighth::ms => now;
            if (i == 0 && ((pattern >= 0.55 && pattern < 0.7) || (pattern > 0.85 && pattern <= 1))) {
                spork ~ runKick();
            } else if (i == 1 && pattern >= 0.3 && pattern < 0.55) {
                spork ~ runKick();
            } else if (i == 2 && pattern >= 0.7) {
                spork ~ runKick();
            }
        }
    }
}
spork ~ kickPattern();

// snare
SndBuf snare => drGain;
me.dir() + "/sounds/snare.wav" => snare.read;
0 => snare.gain;

fun void runSnare() {
    0 => snare.pos;
    1::second => now;
}
fun void snarePattern() {
    while (true) {
        5::ms => now;
        (eighth-5)::ms => now;
        for (0 => int i; i < 3; i++) {
            eighth::ms => now;
            if (i == 1) {
                spork ~ runSnare();
            }
        }
    }
}
spork ~ snarePattern();

// hihat
SndBuf hihat => drGain;
me.dir() + "/sounds/hihat.wav" => hihat.read;
0 => hihat.gain;

fun void runHihat() {
    0 => hihat.pos;
    1::second => now;
}
fun void hihatPattern() {
    while (true) {
        5::ms => now;
        (eighth-5)::ms => now;
        spork ~ runHihat();
        for (0 => int i; i < 3; i++) {
            eighth::ms => now;
            spork ~ runHihat();
        }
    }
}
spork ~ hihatPattern();

// fade in/out to trigger elements
fun void fadeInOut(UGen synth) {
    if (synth.gain() == 0) {
        for (0 => int i; i < 1000; i++) {
            synth.gain(synth.gain()+0.001);
            (eighth/1000)::ms => now;
        }
        1 => synth.gain;
    } else if (synth.gain() == 1) {
        for (0 => int i; i < 1000; i++) {
            synth.gain(synth.gain()-0.001);
            (eighth/1000)::ms => now;
        }
        0 => synth.gain;
    }
}

// constant application of freqMultiplier to lpf freq
fun void adjustLPF() {
    while (true) {
        lpf.freq() => float newFreq;
        freqMultiplier *=> newFreq;
        if (newFreq < 200) {
            200 => newFreq;
        } else if (newFreq > 20000) {
            20000 => newFreq;
        }
        newFreq => lpf.freq;
        if (freqMultiplier != 1) {
            <<< "low pass set to", newFreq, "Hz" >>>;
        }
        5::ms => now;
    }
}
spork ~ adjustLPF();

// infinite event loop
while(true) {
    // wait on HidIn as event
    hi => now;

    // messages received
    while(hi.recv(msg)) {
        // joystick axis motion
        if(msg.isAxisMotion()) {
            if (msg.which != 2 && Std.fabs(msg.axisPosition) > 0.05) { // joysticks
                if (leftJoyMode == 0 && !leftJoyLocked) { // gain mode
                    if (msg.which == 0 && msg.axisPosition < 0) { // left axis, instruments
                        <<< "instrument gain:", msg.axisPosition + 1 >>>;
                        (msg.axisPosition + 1)*inGainDefault => inGain.gain;
                    } else if (msg.which == 1 && msg.axisPosition > 0) { // down axis, drums
                        <<< "drum gain:", 1 - msg.axisPosition >>>;
                        (1 - msg.axisPosition) => drGain.gain;
                    }
                }
                if (leftJoyMode == 1 && !leftJoyLocked) { // reverb mode
                    if (msg.which == 0 && msg.axisPosition > 0) { // right axis, instruments
                        <<< "instrument reverb mix:", msg.axisPosition / 4 >>>;
                        msg.axisPosition / 4 => inRev.mix;
                    } else if (msg.which == 1 && msg.axisPosition < 0) { // up axis, drums
                        <<< "drum reverb mix:", -1*msg.axisPosition / 4 >>>;
                        -1*msg.axisPosition / 4 => drRev.mix;
                    }
                }
                if (rightJoyMode == 0 && !rightJoyLocked) { // arp ADSR mode
                    if (msg.which == 3 && msg.axisPosition > 0) { // right axis, attack
                        <<< "arp attack:", msg.axisPosition >>>;
                        (2*msg.axisPosition*eighth)::ms => arpADSR.attackTime;
                    } else if (msg.which == 4 && msg.axisPosition > 0) { // down axis, decay
                        <<< "arp decay:", -1.75*msg.axisPosition+2 >>>;
                        ((-1.75*msg.axisPosition+2)*eighth)::ms => arpADSR.decayTime;
                    }
                }
                if (rightJoyMode == 1 && !rightJoyLocked) { // delay mode
                    if (msg.which == 3 && msg.axisPosition > 0) { // right axis, instruments
                        <<< "instrument delay mix:", msg.axisPosition / 4 >>>;
                        msg.axisPosition / 4 => inDel.mix;
                    } else if (msg.which == 4 && msg.axisPosition < 0) { // up axis, drums
                        <<< "drum delay mix:", -1*msg.axisPosition / 4 >>>;
                        -1*msg.axisPosition / 4 => drDel.mix;
                    }
                }
            } else if (msg.which == 2) { // triggers
                1 => freqMultiplier;
                if (Math.fabs(msg.axisPosition) > 0.25) { // 0.75 to 1 on left, 1 to 1.25 on right
                    if (msg.axisPosition > 0) {
                        1-(msg.axisPosition-0.25)/15 => freqMultiplier;
                    } else {
                        1-(msg.axisPosition+0.25)/15 => freqMultiplier;
                    }
                }
            }
        }
        
        // joystick button down
        else if(msg.isButtonDown()) {
            eighth::ms - (now % eighth::ms) => now;
            if (msg.which == 0) { // chord change
                currentChord++;
                if (currentChord == 4) {
                    0 => currentChord;
                }
                <<< "chord changed", "" >>>;
            } else if (msg.which == 1) { // arp trigger
                fadeInOut(arp);
                <<< "arp toggled", "" >>>;
            } else if (msg.which == 2) { // bass trigger
                fadeInOut(bass);
                <<< "bass toggled", "" >>>;
            } else if (msg.which == 3) { // chord trigger
                spork ~ fadeInOut(chord0);
                spork ~ fadeInOut(chord1);
                spork ~ fadeInOut(chord2);
                spork ~ fadeInOut(chord3);
                <<< "chords toggled", "" >>>;
            } else if (msg.which == 4) { // kick trigger
                fadeInOut(kick);
                <<< "kick toggled", "" >>>;
            } else if (msg.which == 5) { // snare trigger
                fadeInOut(snare);
                <<< "snare toggled", "" >>>;
            } else if (msg.which == 6 && !tempoChangeWaiting && eighth < 300) { // decrease tempo
                true => tempoChangeWaiting;
                while (true) {
                    eighth::ms - (now % eighth::ms) => now;
                    1::ms => now;
                    if (downbeat) {
                        20 +=> eighth;
                        <<< "eighth note set to", eighth $ int, "ms" >>>;
                        false => tempoChangeWaiting;
                        break;
                    }
                }
            } else if (msg.which == 7 && !tempoChangeWaiting && eighth > 100) { // increase tempo
                true => tempoChangeWaiting;
                while (true) {
                    eighth::ms - (now % eighth::ms) => now;
                    1::ms => now;
                    if (downbeat) {
                        20 -=> eighth;
                        <<< "eighth note set to", eighth $ int, "ms" >>>;
                        false => tempoChangeWaiting;
                        break;
                    }
                }
            } else if (msg.which == 8) { // toggle left joystick lock
                !leftJoyLocked => leftJoyLocked;
                if (leftJoyLocked) {
                    <<< "left joystick locked", "" >>>;
                } else {
                    <<< "left joystick unlocked", "" >>>;    
                }
            } else if (msg.which == 9) { // toggle right joystick lock
                !rightJoyLocked => rightJoyLocked;
                if (rightJoyLocked) {
                    <<< "right joystick locked", "" >>>;
                } else {
                    <<< "right joystick unlocked", "" >>>;    
                }
            } else if (msg.which == 10) { // hihat trigger
                fadeInOut(hihat);
                <<< "hihat toggled", "" >>>;
            }
        }
        
        // joystick hat/POV switch/d-pad motion
        else if(msg.isHatMotion()) {
            if (msg.idata == 8) { // gain mode
                0 => leftJoyMode;
                <<< "left joystick set to gain mode (bottom left quadrant)", "" >>>;
            } else if (msg.idata == 1) { // reverb mode
                1 => leftJoyMode;
                <<< "left joystick set to reverb mode (top right quadrant)", "" >>>;
            } else if (msg.idata == 2) { // arp ADSR mode
                0 => rightJoyMode;
                <<< "right joystick set to arp ADSR mode (bottom right quadrant)", "" >>>;
            } else if (msg.idata == 4) { // delay mode
                1 => rightJoyMode;
                <<< "right joystick set to delay mode (top right quadrant)", "" >>>;
            }
        }
    }
}